'use strict';
// ext. libs
var Q = require('q');
var fs = require('fs');
var phantomjs = require('phantomjs');
var portscanner = require('portscanner');
var spawn = require('child_process').spawn;
var PhantomJSDriver = {

  /**
   * Verbose version of the browser name
   *
   * @property
   * @type string
   * @default PhantomJS
   */

  longName: 'PhantomJS',

  /**
   * Default port of the PhantomJSDriver
   * The port may change, cause the port conflict resultion
   * tool might pick another one, if the default one is blocked
   *
   * @property
   * @type integer
   * @default 9001
   */

  port: 9001,

  /**
   * Default maximum port of the Ghostdriver Server
   * The port is the highest port in the range that can be allocated
   * by the Ghostdriver Server
   *
   * @property maxPort
   * @type integer
   * @default 9091
   */

  maxPort: 9091,

  /**
   * Default host of the PhantomJSDriver
   * The host may be overriden with
   * a user configured value
   *
   * @property
   * @type string
   * @default localhost
   */

  host: 'localhost',

  /**
   * Root path of the PhantomJSDriver
   *
   * @property
   * @type string
   * @default /wd/hub
   */

  path: '/wd/hub',

  /**
   * Default desired capabilities that should be
   * transferred when the browser session gets requested
   *
   * @property desiredCapabilities
   * @type object
   */

  desiredCapabilities: {
    version: phantomjs.version,
    browserName: 'phantomjs'
  },

  /**
   * Driver defaults, what should the driver be able to access.
   *
   * @property driverDefaults
   * @type object
   */

  driverDefaults: {
    viewport: true,
    status: true,
    sessionInfo: true
  },

  /**
   * Child process instance of the PhantomJS browser
   *
   * @property
   * @type null|Object
   */

  spawned: null,

  /**
   * Resolves the driver port
   *
   * @method getPort
   * @return integer
   */

  getPort: function () {
    return this.port;
  },

  /**
   * Resolves the maximum range for the driver port
   *
   * @method getMaxPort
   * @return {integer} port Max WebDriver server port range
   */

  getMaxPort: function () {
    return this.maxPort;
  },

  /**
   * Returns the driver host
   *
   * @method getHost
   * @type string
   */

  getHost: function () {
    return this.host;
  },

  /**
   * Launches PhantomJS, negoatiates a port & checks for a user set binary
   *
   * @method launch
   * @param {object} configuration Browser configuration
   * @param {EventEmitter2} events EventEmitter (Reporter Emitter instance)
   * @param {Internal.Config} config configuration class
   * @return {object} promise Browser promise
   */

  launch: function (configuration, events, config) {
    var deferred = Q.defer();

    // store injected configuration/log event handlers
    this.reporterEvents = events;
    this.configuration = configuration;
    this.config = config;

    // check for a user set port
    var browsers = this.config.get('browsers');
    if (browsers && Array.isArray(browsers)) {
      browsers.forEach(this._checkUserDefinedPorts.bind(this));
    }

    // check if the current port is in use, if so, scan for free ports
    portscanner.findAPortNotInUse(this.getPort(), this.getMaxPort(), this.getHost(), this._checkPorts.bind(this, deferred));
    return deferred.promise;
  },

  /**
   * Kills the PhantomJSDriver processe
   *
   * @method kill
   * @chainable
   */

  kill: function () {
    this.spawned.kill('SIGTERM');
    return this;
  },

  /**
   * Checks if the def. port is blocked & if we need to switch to another port
   * Kicks off the process manager (for closing the opened browsers after the run has been finished)
   * Also starts the chromedriver instance 
   *
   * @method _checkPorts
   * @param {object} deferred Promise
   * @param {null|object} error Error object
   * @param {integer} port Found open port
   * @private
   * @chainable
   */

  _checkPorts: function (deferred, error, port) {
    // check if the port was blocked & if we need to switch to another port
    if (this.port !== port) {
      this.reporterEvents.emit('report:log:system', 'cb-browser-phantomjs: Switching to port: ' + port);
      this.port = port;
    }

    // check the binary
    var binary = this._checkUserDefinedBinary(this.configuration, phantomjs.path);

    // launch the browser process
    this.spawned = spawn(binary, ['--webdriver', this.getPort(), '--ignore-ssl-errors=true']);
    this.spawned.stdout.on('data', this._launch.bind(this, deferred));
    return this;
  },

  /**
   * Checks the data stream from the launched phantom process
   *
   * @method _launch
   * @param {object} deferred Promise
   * @param {buffer} data Console output from Ghostdriver
   * @chainable
   * @private
   */

  _launch: function (deferred, data) {
    var stream = data + '';
    
    // check if ghostdriver could be launched    
    if (stream.search('GhostDriver - Main - running') !== -1) {
      deferred.resolve();
    } else if (stream.search('Could not start Ghost Driver') !== -1) {
      this.reporterEvents.emit('error', 'cb-browser-phantomjs: Could not start Ghost Driver');
      deferred.reject('Could not start Ghost Driver');
      process.exit(127);
    }

    return this;
  },

  /**
   * Process user defined ports
   *
   * @method _checkUserDefinedPorts
   * @param {object} browser Browser configuration
   * @chainable
   * @private
   */

  _checkUserDefinedPorts: function (browser) {
    // check for a single defined port
    if (browser.phantomjs && browser.phantomjs.port) {
      this.port = parseInt(browser.phantomjs.port, 10);
      this.maxPort = this.port + 90;
      this.reporterEvents.emit('report:log:system', 'cb-browser-phantomjs: Switching to user defined port: ' + this.port);
    }

    // check for a port range
    if (browser.phantomjs && browser.phantomjs.portRange && browser.phantomjs.portRange.length === 2) {
      this.port = parseInt(browser.phantomjs.portRange[0], 10);
      this.maxPort = parseInt(browser.phantomjs.portRange[1], 10);
      this.reporterEvents.emit('report:log:system', 'cb-browser-phantomjs: Switching to user defined port(s): ' + this.port + ' -> ' + this.maxPort);
    }

    return this;
  },

  /**
   * Checks if the binary exists,
   * when set manually by the user
   *
   * @method _checkUserDefinedBinary
   * @param {string} binary Path to the browser binary
   * @return {bool|string} Binary path if binary exists, else false
   * @private
   */

  _checkUserDefinedBinary: function (configuration, defaultBinary) {
    var binary = defaultBinary;

    // check if we have a user defined binary
    if (configuration && configuration.binary) {
      binary = configuration.binary;
    }

    // check if we need to replace the users home directory
    if (process.platform === 'darwin' && binary.trim()[0] === '~') {
      binary = binary.replace('~', process.env.HOME);
    }

    // check if the binary exists
    if (!fs.existsSync(binary)) {
      this.reporterEvents.emit('error', 'cb-driver-phantomjs: Binary not found: ' + binary);
      process.exit(127);
      return false;
    }

    return binary;
  }

};

module.exports = PhantomJSDriver;
